// Thai.cpp
// (c) 2004 exeal

#include "StdAfx.h"
#include "Encoder.h"

using namespace Ascension::Encodings;
using namespace std;


DEFINE_ENCODER_CLASS(CPEX_THAI_TIS620, Thai_Tis620);
DEFINE_ENCODER_CLASS(CPEX_ISO8859_11, Iso8859_11);

namespace {
	template<bool bIso8859>
	inline wchar_t ThaiNativeCharToUcs(uchar ch) {
		if(ch < 0x80)						return ch;
		else if(ch == 0xA0)					return bIso8859 ? 0x00A0 : __REPLACEMENT_CHARACTER;
		else if(ch >= 0xA1 && ch <= 0xDA)	return ch + 0x0D60;
		else if(ch >= 0xDF && ch <= 0xFB)	return ch + 0x0D60;
		else								return __REPLACEMENT_CHARACTER;
	}
	template<bool bIso8859>
	inline uchar UcsToThaiNativeChar(wchar_t wch) {
		assert(wch != 0);
		if(wch < 0x0080)						return static_cast<uchar>(wch);
		else if(wch == 0x00A0)					return bIso8859 ? 0xA0 : 0x00;
		else if(wch >= 0x0E01 && wch <= 0x0E3A)	return static_cast<uchar>(wch - 0x0D60);
		else if(wch >= 0x0E3F && wch <= 0x0E5B)	return static_cast<uchar>(wch - 0x0D60);
		else									return 0x00;
	}
	template<bool bIso8859>
	inline size_t UnicodeToThaiNative(
			uchar* pszDest, size_t cchDest, const wchar_t* pwszSrc, size_t cchSrc, bool(*pCallBack)(void)) {
		for(size_t i = 0; i < cchSrc && i < cchDest; ++i) {
			pszDest[i] = UcsToThaiNativeChar<bIso8859>(pwszSrc[i]);
			if(pszDest[i] == 0x00) {
				if(pCallBack == 0 || (*pCallBack)()) {
					pszDest[i] = __DEFAULT_CHAR;
					pCallBack = 0;
				} else
					return 0;
			}
		}
		return min(cchSrc, cchDest);
	}
	template<bool bIso8859>
	inline size_t ThaiNativeToUnicode(
			wchar_t* pwszDest, size_t cchDest, const uchar* pszSrc, size_t cchSrc, bool(*pCallBack)(void)) {
		for(size_t i = 0; i < cchSrc && i < cchDest; ++i) {
			pwszDest[i] = ThaiNativeCharToUcs<false>(pszSrc[i]);
			if(pwszDest[i] == __REPLACEMENT_CHARACTER) {
				if(pCallBack == 0 || (*pCallBack)()) {
					pwszDest[i] = __DEFAULT_CHAR;
					pCallBack = 0;
				} else
					return 0;
			}
		}
		return min(cchSrc, cchDest);
	}
}


// ^C (TIS 620-2533:1990) ///////////////////////////////////////////////

CEncoder_Thai_Tis620::CEncoder_Thai_Tis620() {
}

size_t CEncoder_Thai_Tis620::ConvertFromUnicode(CFU_ARGLIST) {
	CFU_CHECKARGS();
	return UnicodeToThaiNative<false>(reinterpret_cast<uchar*>(pszDest), cchDest, pwszSrc, cchSrc, pCallBack);
}

size_t CEncoder_Thai_Tis620::ConvertToUnicode(CTU_ARGLIST) {
	CTU_CHECKARGS();
	return ThaiNativeToUnicode<false>(pwszDest, cchDest, pszSrc, cchSrc, pCallBack);
}

uchar CEncoder_Thai_Tis620::GetMaxCharacterLength() const {
	return 1;
}


// ^C (ISO-8859-11) /////////////////////////////////////////////////////

CEncoder_Iso8859_11::CEncoder_Iso8859_11() {
}

size_t CEncoder_Iso8859_11::ConvertFromUnicode(CFU_ARGLIST) {
	CFU_CHECKARGS();
	return UnicodeToThaiNative<true>(reinterpret_cast<uchar*>(pszDest), cchDest, pwszSrc, cchSrc, pCallBack);
}

size_t CEncoder_Iso8859_11::ConvertToUnicode(CTU_ARGLIST) {
	CTU_CHECKARGS();
	return ThaiNativeToUnicode<true>(pwszDest, cchDest, pszSrc, cchSrc, pCallBack);
}

uchar CEncoder_Iso8859_11::GetMaxCharacterLength() const {
	return 1;
}

/* [EOF] */