// KeyboardMap.h
// (c) 2003-2004 exeal

#ifndef _KEYBOARD_MAP_H_
#define _KEYBOARD_MAP_H_

#include "AlphaInterfaces.h"
#include "resource.h"
#include "../Manah/Object.h"
#include <sstream>


namespace Alpha {
	///	zL[R[ȟ^
	typedef WORD	VirtualKey;	// typeof(ACCEL::key)
	const VirtualKey	VK_NULL = 0;	///< ȃL[

	///	zL[̏CL[
	typedef uchar	KeyModifier;
	const KeyModifier	KM_SHIFT	= 0x01;	///< Shift L[
	const KeyModifier	KM_CTRL		= 0x02;	///< Ctrl L[
	const KeyModifier	KM_ALT		= 0x04;	///< Alt L[

	///	L[oCh̊Ǘ
	class CKeyboardMap : virtual public ISerializable {
		// RXgN^
	public:
		CKeyboardMap();
		~CKeyboardMap();

		// \bh
	public:
		void			Assign(CommandId id, VirtualKey key, KeyModifier modifiers) throw(std::out_of_range);
		void			Assign(CommandId id, VirtualKey firstKey, KeyModifier firstModifiers,
							VirtualKey secondKey, KeyModifier secondModifiers) throw(std::out_of_range);
		void			Clear();
		CommandId		GetCommandID(VirtualKey key, KeyModifier modifiers) const throw(std::out_of_range);
		CommandId		GetCommandID(VirtualKey firstKey, KeyModifier firstModifiers,
							VirtualKey secondKey, KeyModifier secondModifiers) const throw(std::out_of_range);
		std::wstring	GetKeyString(CommandId id, bool bShortName) const;

		static const wchar_t*	GetKeyName(VirtualKey key, bool bShortName) throw(std::out_of_range);
		static std::wstring		GetStrokeString(VirtualKey key, KeyModifier modifiers, bool bShortName) throw(std::out_of_range);
		static std::wstring		GetStrokeString(VirtualKey firstKey, KeyModifier firstModifiers,
									VirtualKey secondKey, KeyModifier secondModifiers, bool bShortName) throw(std::out_of_range);

		// ISerializable \bh
		bool	IsDirty() const;
		bool	Load(const wchar_t* pwszPathName);
		bool	Save(const wchar_t* pwszPathName);

		// f[^o
	private:
		struct T1stKeyMap {
			CommandId	id;				// L[V[PX1Xg[NŏIꍇ̓R}h ID
			CommandId**	pp2ndKeyMap;	// 2Xg[Ngꍇ 8 * 0x0100 ̔z
										// R}h蓖ĂĂȂꍇ͂̃o0
		};
		T1stKeyMap	m_firstKeyMaps[8][0x0100];	// (CL[) * (zL[R[h) ŃR}h ID ͑2L[gݍ킹i[B
												// 0łΉo^ĂȂ
		bool		m_bDirty;
	};


	namespace {
		// L[̖O (蓖ĕs\ȃL[ null)
		const wchar_t*	longKeyNames[0x0100] = {
		/* 0x00 */	0,	0,	0,	0,	0,	0,	0,	0,
					L"BackSpace",	L"Tab",	0,	0,	L"Delete",	L"Enter",	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					L"Esc",	0,	0,	0,	0,	0,	0,	0,
		/* 0x20 */	L"Space",	L"PageUp",	L"PageDown",	L"End",	L"Home",	L"Left",	L"Top",	L"Right",
					L"Down",	0,	0,	0,	0,	L"Insert",	0,	0,
					L"0",	L"1",	L"2",	L"3",	L"4",	L"5",	L"6",	L"7",
					L"8",	L"9",	0,	0,	0,	0,	0,	0,
		/* 0x40 */	0,	L"A",	L"B",	L"C",	L"D",	L"E",	L"F",	L"G",
					L"H",	L"I",	L"J",	L"K",	L"L",	L"M",	L"N",	L"O",
					L"P",	L"Q",	L"R",	L"S",	L"T",	L"U",	L"V",	L"W",
					L"X",	L"Y",	L"Z",	0,	0,	0,	0,	0,
		/* 0x60 */	0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	L"numpad *",	L"numpad +",	L"|",	L"numpad -",	L"numpad .",	L"numpad /",
					L"F1",	L"F2",	L"F3",	L"F4",	L"F5",	L"F6",	L"F7",	L"F8",
					L"F9",	L"F10",	L"F11",	L"F12",	L"F13",	L"F14",	L"F15",	L"F16",
		/* 0x80 */	L"F17",	L"F18",	L"F19",	L"F20",	L"F21",	L"F22",	L"F23",	L"F24",
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
		/* 0xA0 */	0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	L":",	L";",	L",",	L"-",	L".",	L"/",
		/* 0xC0 */	L"@",	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	L"[",	L"\\",	L"]",	L"^",	0,
		/* 0xE0 */	0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0
		};
		const wchar_t*	shortKeyNames[0x0100] = {
		/* 0x00 */	0,	0,	0,	0,	0,	0,	0,	0,
					L"BS",	L"TAB",	0,	0,	L"Delete",	L"RET",	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					L"ESC",	0,	0,	0,	0,	0,	0,	0,
		/* 0x20 */	L"SPC",	L"PageUp",	L"PageDown",	L"End",	L"Home",	L"Left",	L"Up",	L"Right",
					L"Down",	0,	0,	0,	0,	L"Insert",	0,	0,
					L"0",	L"1",	L"2",	L"3",	L"4",	L"5",	L"6",	L"7",
					L"8",	L"9",	0,	0,	0,	0,	0,	0,
		/* 0x40 */	0,	L"a",	L"b",	L"c",	L"d",	L"e",	L"f",	L"g",
					L"h",	L"i",	L"j",	L"k",	L"l",	L"m",	L"n",	L"o",
					L"p",	L"q",	L"r",	L"s",	L"t",	L"u",	L"v",	L"w",
					L"x",	L"y",	L"z",	0,	0,	0,	0,	0,
		/* 0x60 */	0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	L"NUMPAD *",	L"NUMPAD +",	L"|",	L"NUMPAD -",	L"NUMPAD .",	L"NUMPAD /",
					L"F1",	L"F2",	L"F3",	L"F4",	L"F5",	L"F6",	L"F7",	L"F8",
					L"F9",	L"F10",	L"F11",	L"F12",	L"F13",	L"F14",	L"F15",	L"F16",
		/* 0x80 */	L"F17",	L"F18",	L"F19",	L"F20",	L"F21",	L"F22",	L"F23",	L"F24",
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
		/* 0xA0 */	0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	L":",	L";",	L",",	L"-",	L".",	L"/",
		/* 0xC0 */	L"@",	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	L"[",	L"\\",	L"]",	L"^",	0,
		/* 0xE0 */	0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0,
					0,	0,	0,	0,	0,	0,	0,	0
		};
	}


	/**
	 *	L[gݍ킹R}h ID 擾
	 *	@param key				zL[
	 *	@param modifier			CL[
	 *	@return					蓖ĂĂR}h IDB蓖ĂĂȂƂ0B
	 *							2Xg[N1Xg[Nڂł CMD_SPECIAL_WAITFOR2NDKEYS
	 *	@exception out_of_range	<var>key</var> sȂƂX[
	 */
	inline CommandId CKeyboardMap::GetCommandID(VirtualKey key, KeyModifier modifiers) const throw(std::out_of_range) {
		if(key >= 0x0100)
			throw std::out_of_range("The first argument is invalid as virutal key code.");
		return m_firstKeyMaps[modifiers][key].id;
	}

	/**
	 *	2Xg[ÑL[gݍ킹R}h ID 擾
	 *	@param firstKey 		1̉zL[
	 *	@param firstModifiers	1̏CL[
	 *	@param secondKey 		2̉zL[
	 *	@param secondModifiers	2̏CL[
	 *	@return					蓖ĂĂR}h IDB蓖ĂĂȂƂ0
	 *	@exception out_of_range	<var>firstKey</var> A<var>secondKey</var> ̂ꂩsȂƂX[
	 */
	inline CommandId CKeyboardMap::GetCommandID(VirtualKey firstKey,
			KeyModifier firstModifiers, VirtualKey secondKey, KeyModifier secondModifiers) const throw(std::out_of_range) {
		if(firstKey >= 0x0100 || secondKey >= 0x0100)
			throw std::out_of_range("The first or third argument is invalid as virutal key code.");
		const T1stKeyMap&	firstKeyMap = m_firstKeyMaps[firstModifiers][firstKey];
		if(firstKeyMap.id != CMD_SPECIAL_WAITFOR2NDKEYS)
			return 0;
		assert(firstKeyMap.pp2ndKeyMap != 0);
		return (firstKeyMap.pp2ndKeyMap[secondModifiers] == 0) ? 0 : firstKeyMap.pp2ndKeyMap[secondModifiers][secondKey];
	}

	/**
	 *	L[̖OԂ
	 *	@param key				zL[
	 *	@param bShortName		Z`̖O擾ꍇ true
	 *	@return					L[BΉL[ꍇ null
	 *	@exception out_of_range	<var>key</var> sȂƂX[
	 */
	inline const wchar_t* CKeyboardMap::GetKeyName(VirtualKey key, bool bShortName) throw(std::out_of_range) {
		if(key >= 0x100)
			throw std::out_of_range("The first argument is invalid as virutal key code.");
		return bShortName ? shortKeyNames[key] : longKeyNames[key];
	}

	///	L[Xg[N\镶Ԃ
	inline std::wstring CKeyboardMap::GetStrokeString(VirtualKey key,
			KeyModifier modifiers, bool bShortName) throw(std::out_of_range) {
		if(modifiers == 0)
			return bShortName ? shortKeyNames[key] : longKeyNames[key];
		std::wostringstream	ss;

		if(toBoolean(modifiers & KM_CTRL))	ss << (bShortName ? L"C-" : L"Ctrl+");
		if(toBoolean(modifiers & KM_SHIFT))	ss << (bShortName ? L"S-" : L"Shift+");
		if(toBoolean(modifiers & KM_ALT))	ss << (bShortName ? L"M-" : L"Alt+");
		ss << (bShortName ? shortKeyNames[key] : longKeyNames[key]);
		return ss.str();
	}

	///	L[Xg[N\镶Ԃ
	inline std::wstring CKeyboardMap::GetStrokeString(VirtualKey firstKey, KeyModifier firstModifiers,
			VirtualKey secondKey, KeyModifier secondModifiers, bool bShortName) throw(std::out_of_range) {
		return GetStrokeString(firstKey, firstModifiers, bShortName)
			+ L" " + GetStrokeString(secondKey, secondModifiers, bShortName);
	}

	///	@see	ISerializable::IsDirty
	inline bool CKeyboardMap::IsDirty() const {
		return m_bDirty;
	}
}

#endif /* _KEYBOARD_MAP_H_ */

/* [EOF] */